@@ -1,12 +1,13 @@
 # coding=utf-8
 #
-# This file is part of Hypothesis (https://github.com/DRMacIver/hypothesis)
+# This file is part of Hypothesis, which may be found at
+# https://github.com/HypothesisWorks/hypothesis-python
 #
 # Most of this work is copyright (C) 2013-2015 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
-# https://github.com/DRMacIver/hypothesis/blob/master/CONTRIBUTING.rst for a
-# full list of people who may hold copyright, and consult the git log if you
-# need to determine who owns an individual contribution.
+# CONTRIBUTING.rst for a full list of people who may hold copyright, and
+# consult the git log if you need to determine who owns an individual
+# contribution.
 #
 # This Source Code Form is subject to the terms of the Mozilla Public License,
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
@@ -32,13 +33,13 @@ from hypothesis.errors import Flaky, Timeout, NoSuchExample, \
     UnsatisfiedAssumption, HypothesisDeprecationWarning
 from hypothesis.control import BuildContext
 from hypothesis._settings import settings as Settings
-from hypothesis._settings import Verbosity
+from hypothesis._settings import Phase, Verbosity, HealthCheck
 from hypothesis.executors import new_style_executor, \
     default_new_style_executor
 from hypothesis.reporting import report, verbose_report, current_verbosity
 from hypothesis.internal.compat import getargspec, str_to_bytes
-from hypothesis.internal.reflection import arg_string, impersonate, \
-    copy_argspec, function_digest, fully_qualified_name, \
+from hypothesis.internal.reflection import nicerepr, arg_string, \
+    impersonate, copy_argspec, function_digest, fully_qualified_name, \
     convert_positional_arguments, get_pretty_function_description
 from hypothesis.searchstrategy.strategies import SearchStrategy
 
@@ -120,6 +121,18 @@ def seed(seed):
     return accept
 
 
+class WithRunner(SearchStrategy):
+
+    def __init__(self, base, runner):
+        assert runner is not None
+        self.base = base
+        self.runner = runner
+
+    def do_draw(self, data):
+        data.hypothesis_runner = self.runner
+        return self.base.do_draw(data)
+
+
 def given(*generator_arguments, **generator_kwargs):
     """A decorator for turning a test function that accepts arguments into a
     randomized test.
@@ -238,6 +251,8 @@ def given(*generator_arguments, **generator_kwargs):
                     ))
                 else:
                     example_kwargs = example.kwargs
+                if Phase.explicit not in settings.phases:
+                    continue
                 example_kwargs.update(kwargs)
                 # Note: Test may mutate arguments and we can't rerun explicit
                 # examples, so we have to calculate the failure message at this
@@ -268,14 +283,23 @@ def given(*generator_arguments, **generator_kwargs):
                 )
             )
 
-            def fail_health_check(message):
+            def fail_health_check(message, label):
+                if label in settings.suppress_health_check:
+                    return
                 message += (
                     '\nSee http://hypothesis.readthedocs.org/en/latest/health'
-                    'checks.html for more information about this.'
+                    'checks.html for more information about this. '
                 )
+                message += (
+                    'If you want to disable just this health check, add %s '
+                    'to the suppress_health_check settings for this test.'
+                ) % (label,)
                 raise FailedHealthCheck(message)
 
             search_strategy = given_specifier
+            if selfy is not None:
+                search_strategy = WithRunner(search_strategy, selfy)
+
             search_strategy.validate()
 
             perform_health_check = settings.perform_health_check
@@ -283,6 +307,11 @@ def given(*generator_arguments, **generator_kwargs):
 
             from hypothesis.internal.conjecture.data import TestData, Status, \
                 StopTest
+            if not (
+                Phase.reuse in settings.phases or
+                Phase.generate in settings.phases
+            ):
+                return
 
             if perform_health_check:
                 initial_state = getglobalrandomstate()
@@ -332,6 +361,8 @@ def given(*generator_arguments, **generator_kwargs):
                         else:
                             assert data.status == Status.OVERRUN
                             overruns += 1
+                    except InvalidArgument:
+                        raise
                     except Exception:
                         report(traceback.format_exc())
                         if test_runner is default_new_style_executor:
@@ -340,8 +371,9 @@ def given(*generator_arguments, **generator_kwargs):
                                 'generation in initial health check. '
                                 'This indicates a bug in the strategy. '
                                 'This could either be a Hypothesis bug or '
-                                "an error in a function yo've passed to "
-                                'it to construct your data.'
+                                "an error in a function you've passed to "
+                                'it to construct your data.',
+                                HealthCheck.exception_in_generation,
                             )
                         else:
                             fail_health_check(
@@ -353,7 +385,8 @@ def given(*generator_arguments, **generator_kwargs):
                                 'it to construct your data. Additionally, '
                                 'you have a custom executor, which means '
                                 'that this could be your executor failing '
-                                'to handle a function which returns None. '
+                                'to handle a function which returns None. ',
+                                HealthCheck.exception_in_generation,
                             )
                 if overruns >= 20 or (
                     not count and overruns > 0
@@ -366,7 +399,7 @@ def given(*generator_arguments, **generator_kwargs):
                         'max_size parameters on your collections and turning '
                         'max_leaves down on recursive() calls.') % (
                         overruns, count
-                    ))
+                    ), HealthCheck.data_too_large)
                 if filtered_draws >= 50 or (
                     not count and filtered_draws > 0
                 ):
@@ -379,7 +412,7 @@ def given(*generator_arguments, **generator_kwargs):
                         'strategy to filter less. This can also be caused by '
                         'a low max_leaves parameter in recursive() calls') % (
                         filtered_draws, count
-                    ))
+                    ), HealthCheck.filter_too_much)
                 runtime = time.time() - start
                 if runtime > 1.0 or count < 10:
                     fail_health_check((
@@ -388,7 +421,9 @@ def given(*generator_arguments, **generator_kwargs):
                         'and %d exceeded maximum size). Try decreasing '
                         "size of the data you're generating (with e.g."
                         'average_size or max_leaves parameters).'
-                    ) % (count, runtime, filtered_draws, overruns))
+                    ) % (count, runtime, filtered_draws, overruns),
+                        HealthCheck.too_slow,
+                    )
                 if getglobalrandomstate() != initial_state:
                     fail_health_check(
                         'Data generation depends on global random module. '
@@ -398,7 +433,7 @@ def given(*generator_arguments, **generator_kwargs):
                         'randoms() from hypothesis.strategies to get an '
                         'instance of Random you can use. Alternatively, you '
                         'can use the random_module() strategy to explicitly '
-                        'seed the random module.'
+                        'seed the random module.', HealthCheck.random_module,
                     )
             last_exception = [None]
             repr_for_last_exception = [None]
@@ -415,10 +450,10 @@ def given(*generator_arguments, **generator_kwargs):
                         search_strategy, test,
                     ))
                     if result is not None and settings.perform_health_check:
-                        raise FailedHealthCheck((
+                        fail_health_check((
                             'Tests run under @given should return None, but '
                             '%s returned %r instead.'
-                        ) % (test.__name__, result), settings)
+                        ) % (test.__name__, result), HealthCheck.return_value)
                     return False
                 except UnsatisfiedAssumption:
                     data.mark_invalid()
@@ -442,7 +477,9 @@ def given(*generator_arguments, **generator_kwargs):
                             'consider using the randoms() strategy from '
                             'hypothesis.strategies instead. Alternatively, '
                             'you can use the random_module() strategy to '
-                            'explicitly seed the random module.')
+                            'explicitly seed the random module.',
+                            HealthCheck.random_module,
+                        )
 
             from hypothesis.internal.conjecture.engine import TestRunner
 
@@ -460,6 +497,8 @@ def given(*generator_arguments, **generator_kwargs):
                 settings.timeout > 0 and
                 run_time >= settings.timeout
             )
+            if runner.last_data is None:
+                return
             if runner.last_data.status == Status.INTERESTING:
                 falsifying_example = runner.last_data.buffer
                 if settings.database is not None:
@@ -584,16 +623,16 @@ def find(specifier, condition, settings=None, random=None, database_key=None):
         if settings.verbosity == Verbosity.verbose:
             if not successful_examples[0]:
                 report(lambda: u'Trying example %s' % (
-                    repr(result),
+                    nicerepr(result),
                 ))
             elif success:
                 if successful_examples[0] == 1:
                     report(lambda: u'Found satisfying example %s' % (
-                        repr(result),
+                        nicerepr(result),
                     ))
                 else:
                     report(lambda: u'Shrunk example to %s' % (
-                        repr(result),
+                        nicerepr(result),
                     ))
                 last_data[0] = data
         if success and not data.frozen:
